/*
    Copyright (C) 2008    Dmitry Ivanov <vonami@gmail.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "feedlistmodel.h"
#include "../libkrss/tagscollectionattribute.h"
#include "../libkrss/feedpropertiescollectionattribute.h"
#include "../libkrss/subscriptionlabelscollectionattribute.h"

#include <KDebug>
#include <KLocale>

#include <QHashIterator>

FeedlistModel::FeedlistModel( QObject *parent )
        : QAbstractItemModel( parent ), m_sourceModel(0),
          m_allFeedsTag( i18n( "All feeds" ) )
{
        m_rootNode = new TreeNode( TreeNode::RootNode, "Subscriptions", 0 );
        TreeNode *tagNode = new TreeNode( TreeNode::TagNode, m_allFeedsTag, m_rootNode );
        m_rootNode->appendChild( tagNode );
        m_tagNodes[ m_allFeedsTag ] = tagNode;
}

FeedlistModel::~FeedlistModel()
{
        delete m_rootNode;
}

void FeedlistModel::setSourceModel( Akonadi::CollectionModel *model )
{
        m_sourceModel = new Akonadi::CollectionFilterProxyModel( this );
        m_sourceModel->addMimeTypeFilter( "application/rss+xml" );
        m_sourceModel->setSourceModel( model );

        connect( m_sourceModel, SIGNAL( rowsInserted( const QModelIndex&, int, int ) ),
                 this, SLOT( slotRowsInserted( const QModelIndex&, int, int) ) );
        connect( m_sourceModel, SIGNAL( rowsRemoved( const QModelIndex&, int, int ) ),
                 this, SLOT( slotRowsRemoved( const QModelIndex&, int, int) ) );
        connect( m_sourceModel, SIGNAL( dataChanged( const QModelIndex&, const QModelIndex& ) ),
                 this, SLOT( slotDataChanged( const QModelIndex&, const QModelIndex& ) ) );
}

Akonadi::Collection FeedlistModel::rootCollection() const
{
        QModelIndex subs = m_sourceModel->index( 0, 0, QModelIndex() );
        return m_sourceModel->data( subs, Akonadi::CollectionModel::CollectionRole ).value<Akonadi::Collection>();
}

Akonadi::Collection FeedlistModel::collectionForId( const Akonadi::Collection::Id &colid ) const
{
        QModelIndex subs = m_sourceModel->index( 0, 0, QModelIndex() );
        QModelIndex sourceIndex = m_sourceModel->index( m_collections.indexOf( colid ), 0, subs );
        return m_sourceModel->data( sourceIndex, Akonadi::CollectionModel::CollectionRole ).value<Akonadi::Collection>();
}

QStringList FeedlistModel::allTags() const
{
        return m_tagNodes.keys();
}

QList<Akonadi::Collection::Id> FeedlistModel::allCollections() const
{
        return m_collections;
}

QString FeedlistModel::xmlUrl( const QModelIndex &index ) const
{
        return data( index, FeedlistModel::XmlUrlRole ).toString();
}

QString FeedlistModel::htmlUrl( const QModelIndex &index ) const
{
        return data( index, FeedlistModel::HtmlUrlRole ).toString();
}

QString FeedlistModel::description( const QModelIndex &index ) const
{
        return data( index, FeedlistModel::DescriptionRole ).toString();
}

QStringList FeedlistModel::tags( const QModelIndex &index ) const
{
        return data( index, FeedlistModel::TagsRole ).toStringList();
}

QStringList FeedlistModel::subscriptionLabels( const QModelIndex &index ) const
{
        return data( index, FeedlistModel::SubscriptionLabelsRole ).toStringList();
}

QVariant FeedlistModel::headerData( int section, Qt::Orientation orientation, int role ) const
{
        if ( section == 0 && orientation == Qt::Horizontal && role == Qt::DisplayRole )
                return i18n( "Feeds" );

        return QVariant();
}

int FeedlistModel::columnCount( const QModelIndex &parent ) const
{
        Q_UNUSED( parent );
        return 1;
}

int FeedlistModel::rowCount( const QModelIndex &parent ) const
{
        if ( !parent.isValid() )
                return m_rootNode->count();

        TreeNode *node = static_cast<TreeNode*>( parent.internalPointer() );
        return node->count();
}

QVariant FeedlistModel::data(const QModelIndex &index, int role ) const
{
        if ( !index.isValid() )
                return QVariant();

        // tag node
        TreeNode *node = static_cast<TreeNode*>( index.internalPointer() );
        if ( node->tier() == TreeNode::TagNode ) {
                switch ( role ) {
                        case Qt::DisplayRole:          return node->name();
                        case FeedlistModel::TierRole:  return FeedlistModel::TagTier;
                        default:                       return QVariant();
                }
        }

        // feed node
        if ( node->tier() == TreeNode::FeedNode ) {

                if ( role == FeedlistModel::TierRole ) {
                        return FeedlistModel::FeedTier;
                }

                QModelIndex subs = m_sourceModel->index( 0, 0, QModelIndex() );
                int sourceRow = m_collections.indexOf( node->collection() );
                QModelIndex sourceIndex = m_sourceModel->index( sourceRow, 0, subs );

                if ( role == Qt::DisplayRole || role == FeedlistModel::XmlUrlRole ||
                     role == FeedlistModel::HtmlUrlRole || role == FeedlistModel::DescriptionRole ) {
                        const Akonadi::Collection col = m_sourceModel->data( sourceIndex,
                                                                             Akonadi::CollectionModel::CollectionRole ).
                                                                             value<Akonadi::Collection>();

                        FeedPropertiesCollectionAttribute *props;
                        if ( col.hasAttribute( "FeedProperties" ) )
                                props = col.attribute<FeedPropertiesCollectionAttribute>();
                        else
                                return QString();

                        switch ( role ) {
                                case Qt::DisplayRole:                 return props->name();
                                case FeedlistModel::XmlUrlRole:       return props->xmlUrl();
                                case FeedlistModel::HtmlUrlRole:      return props->htmlUrl();
                                case FeedlistModel::DescriptionRole:  return props->description();
                                default:                              return QString();
                        }
                }
                else if ( role == FeedlistModel::TagsRole ) {
                        const Akonadi::Collection col = m_sourceModel->data( sourceIndex,
                                                                             Akonadi::CollectionModel::CollectionRole ).
                                                                             value<Akonadi::Collection>();

                        if ( col.hasAttribute( "Tags" ) )
                                return col.attribute<TagsCollectionAttribute>()->tags();
                        else
                                return QStringList();
                }
                else if ( role == FeedlistModel::SubscriptionLabelsRole ) {
                        const Akonadi::Collection col = m_sourceModel->data( sourceIndex,
                                                                             Akonadi::CollectionModel::CollectionRole ).
                                                                             value<Akonadi::Collection>();

                        if ( col.hasAttribute( "SubscriptionLabels" ) )
                                return col.attribute<SubscriptionLabelsCollectionAttribute>()->subscriptionLabels();
                        else
                                return QStringList();
                }
                else
                        return m_sourceModel->data( sourceIndex, role );
        }

        return QVariant();
}

QModelIndex FeedlistModel::index( int row, int column, const QModelIndex &parent ) const
{
        if ( row < 0 || column < 0 )
                return QModelIndex();

        // create index for a tag node
        if ( !parent.isValid() ) {
                if ( row < m_rootNode->count() && column == 0 ) {
                        return createIndex( row, 0, m_rootNode->childAt( row ) );
                }
                else {
                        return QModelIndex();
                }
        }

        TreeNode *node = static_cast<TreeNode*>( parent.internalPointer() );
        if ( row < node->count() && column == 0 ) {
                TreeNode *child = node->childAt( row );
                return createIndex( row, 0, child );
        }
        else {
                return QModelIndex();
        }
}

QModelIndex FeedlistModel::parent( const QModelIndex &index ) const
{
        if ( !index.isValid() )
                return QModelIndex();

        TreeNode *node = static_cast<TreeNode*>( index.internalPointer() );

        if ( node->tier() == TreeNode::TagNode ) {
                return QModelIndex();
        }
        else if ( node->tier() == TreeNode::FeedNode ) {
                TreeNode *parentNode = node->parent();
                return createIndex( m_rootNode->childRow( parentNode ), 0, parentNode );
        }

        return QModelIndex();
}

void FeedlistModel::slotDataChanged( const QModelIndex &start, const QModelIndex &end )
{
        QModelIndex subs = m_sourceModel->index( 0, 0, QModelIndex() );
        if ( start.parent() != end.parent() || start.parent() != subs ) {
                return;
        }

        // FIXME: iterate over the indexes
        if ( start != end ) {
                return;
        }

        Akonadi::Collection col = m_sourceModel->data( start, Akonadi::CollectionModel::CollectionRole ).
                                                     value<Akonadi::Collection>();
        QStringList newTags;
        if ( col.hasAttribute( "Tags" ) ) {
                newTags = col.attribute<TagsCollectionAttribute>()->tags();
                if ( newTags.isEmpty() ) {
                       newTags << m_allFeedsTag;
                }
        }
        else {
                newTags << m_allFeedsTag;
        }
        QStringList oldTags = lookupTags( col.id() );
        kDebug() << "New tags:" << newTags;
        kDebug() << "Old tags:" << oldTags;
        QSet<QString> common = oldTags.toSet().intersect( newTags.toSet() );
        QSet<QString> toRemove = oldTags.toSet().subtract( newTags.toSet() );
        QSet<QString> toAdd = newTags.toSet().subtract( oldTags.toSet() );
        kDebug() << "common:" << common;
        kDebug() << "to remove:" << toRemove;
        kDebug() << "to add:" << toAdd;

        // check whether this feed has no tags now
        if ( common.count() == 0 && toAdd.count() == 0 ) {
                toAdd << m_allFeedsTag;
        }

        // updated the feed nodes under the 'common' tags
        Q_FOREACH( const QString &tag, common ) {
                TreeNode *feedNode = lookupFeedNode( col.id(), tag );
                QModelIndex feedIndex = createIndex( feedNode->row(), 0, feedNode );
                emit dataChanged( feedIndex, feedIndex );

                // HACK: update the tag node to trigger updates
                // in FeedlistFilterProxyModel
                emit dataChanged( feedIndex.parent(), feedIndex.parent() );
        }

        // remove the feed nodes from the 'toRemove' tags
        Q_FOREACH( const QString &tag, toRemove ) {
                TreeNode *feedNode = lookupFeedNode( col.id(), tag );
                TreeNode *tagNode = feedNode->parent();
                int tagrow = tagNode->row();
                QModelIndex tagIndex = createIndex( tagrow, 0, tagNode );
                int feedrow = feedNode->row();
                beginRemoveRows( tagIndex, feedrow, feedrow );
                tagNode->removeChildAt( feedrow );
                m_feedNodes[ col.id() ].removeAll( feedNode );
                endRemoveRows();

                // HACK: update the tag node to trigger updates
                // in FeedlistFilterProxyModel
                emit dataChanged( tagIndex, tagIndex );

                // if there is no feed nodes under this tag anymore remove it
                if ( tagNode->count() == 0 && tagNode->name() != m_allFeedsTag ) {
                        beginRemoveRows( QModelIndex(), tagrow, tagrow );
                        m_rootNode->removeChildAt( tagrow );
                        m_tagNodes.remove( tag );
                        endRemoveRows();
                }
        }

        // add new feed nodes under the 'toAdd' tags
        Q_FOREACH( const QString &tag, toAdd ) {
                TreeNode *tagNode;
                // if there is no such a tag yet add a new tag node
                if ( !m_tagNodes.contains( tag ) ) {
                        int tagrow = m_rootNode->count();
                        beginInsertRows( QModelIndex(), tagrow, tagrow );
                        tagNode = new TreeNode( TreeNode::TagNode, tag, m_rootNode );
                        m_rootNode->appendChild( tagNode );
                        m_tagNodes[ tag ] = tagNode;
                        endInsertRows();
                        //dump();
                }
                else {
                        tagNode = m_tagNodes[ tag ];
                }

                QModelIndex tagIndex = createIndex( tagNode->row(), 0, tagNode );
                int feedrow = tagNode->count();
                beginInsertRows( tagIndex, feedrow, feedrow );
                TreeNode *feedNode = new TreeNode( TreeNode::FeedNode, col.id(), tagNode );
                tagNode->appendChild( feedNode );
                m_feedNodes[ col.id() ].append( feedNode );
                endInsertRows();

                // HACK: update the tag node to trigger updates
                // in FeedlistFilterProxyModel
                emit dataChanged( tagIndex, tagIndex );
        }

        //dump();
}

void FeedlistModel::slotRowsInserted( const QModelIndex &parent, int start, int end )
{
        // a root node has been added
        if ( !parent.isValid() ) {
                QModelIndex rootIndex = m_sourceModel->index( 0, 0, parent );
                int count = m_sourceModel->rowCount( rootIndex );
                for ( int i = 0; i < count; ++i ) {
                        QModelIndex child = m_sourceModel->index( i, 0, rootIndex );
                        addFeedNode( child );
                }
        }
        // a feed node has been added
        else if ( parent == m_sourceModel->index( 0, 0, QModelIndex() ) ) {
                for ( int i = start; i <= end; ++i ) {
                        QModelIndex child = m_sourceModel->index( i, 0, parent );
                        addFeedNode( child );
                }
        }

        //dump();
}

void FeedlistModel::slotRowsRemoved( const QModelIndex &parent, int start, int end )
{
        // FIXME: this slot actually should be more generic
        Q_UNUSED( end )

        QModelIndex subs = m_sourceModel->index( 0, 0, QModelIndex() );
        if ( parent != subs ) {
                return;
        }

        Akonadi::Collection::Id col = m_collections[ start ];
        QList<TreeNode*> feedNodes = m_feedNodes[ col ];

        // remove the feed nodes
        Q_FOREACH( TreeNode *feedNode, feedNodes ) {
                TreeNode *tagNode = feedNode->parent();
                int tagrow = tagNode->row();
                QModelIndex tagIndex = createIndex( tagrow, 0, tagNode );
                int feedrow = feedNode->row();
                beginRemoveRows( tagIndex, feedrow, feedrow );
                tagNode->removeChildAt( feedrow );
                m_feedNodes[ col ].removeAll( feedNode );
                endRemoveRows();

                // HACK: update the tag node to trigger updates
                // in FeedlistFilterProxyModel
                emit dataChanged( tagIndex, tagIndex );

                // if there is no feed nodes under this tag anymore remove it
                if ( tagNode->count() == 0 && tagNode->name() != m_allFeedsTag ) {
                        beginRemoveRows( QModelIndex(), tagrow, tagrow );
                        m_tagNodes.remove( tagNode->name() );
                        m_rootNode->removeChildAt( tagrow );
                        endRemoveRows();
                }
        }

        m_collections.removeAll( col );
        //dump();
}

void FeedlistModel::dump() const
{
        kDebug() << "All collections" << m_collections;

        kDebug() << "Feed nodes by collections:";
        QHashIterator<Akonadi::Collection::Id, QList<TreeNode*> > it_col( m_feedNodes );
        while ( it_col.hasNext() ) {
                it_col.next();
                kDebug() << "Col:" << it_col.key() << ", feed nodes:" << it_col.value();
        }

        kDebug() << "Tag nodes by tags:";
        QHashIterator<QString, TreeNode*> it_tag( m_tagNodes );
        while ( it_tag.hasNext() ) {
                it_tag.next();
                kDebug() << "Tag:" << it_tag.key() << ", tag nodes: " << it_tag.value();
        }

        m_rootNode->dump();
}

QList<QModelIndex> FeedlistModel::feedIndexes( const Akonadi::Collection::Id &col ) const
{
        QList<QModelIndex> feedIndexesList;

        QList<TreeNode*> feedNodes = m_feedNodes[ col ];
        Q_FOREACH( TreeNode *feedNode, feedNodes ) {
                feedIndexesList << createIndex ( feedNode->row(), 0, feedNode );
        }

        return feedIndexesList;
}


QStringList FeedlistModel::lookupTags( const Akonadi::Collection::Id &id ) const
{
        QList<TreeNode*> feedNodes = m_feedNodes[ id ];
        QStringList tags;
        Q_FOREACH( const TreeNode *feedNode, feedNodes ) {
                tags << feedNode->parent()->name();
        }

        return tags;
}

TreeNode* FeedlistModel::lookupFeedNode( const Akonadi::Collection::Id &id, const QString &tag ) const
{
        TreeNode *tagNode = m_tagNodes[ tag ];
        QList<TreeNode*> feedNodes = m_feedNodes[ id ];
        Q_FOREACH( TreeNode *feedNode, feedNodes ) {
                if ( feedNode->parent() == tagNode )
                        return feedNode;
        }

        return 0;
}

void FeedlistModel::addFeedNode( const QModelIndex &sourceIndex )
{
        // FIXME: doesn't support inserting a new node between existing ones
        // but akonadi seem to _append_ a new node so that's not going
        // to be a problem for now
        const Akonadi::Collection col = m_sourceModel->data( sourceIndex, Akonadi::CollectionModel::CollectionRole ).
                                                            value<Akonadi::Collection>();
        QStringList feedTags;
        if ( col.hasAttribute( "Tags" ) ) {
                feedTags = col.attribute<TagsCollectionAttribute>()->tags();
                if ( feedTags.isEmpty() ) {
                        feedTags << m_allFeedsTag;
                }
        }
        else {
                feedTags << m_allFeedsTag;
        }

        m_collections << col.id();
        Q_FOREACH( const QString &tag, feedTags ) {
                TreeNode *tagNode;
                if ( m_tagNodes.contains( tag ) ) {
                        tagNode = m_tagNodes[ tag ];
                }
                else {
                        int tagrow = m_rootNode->count();
                        beginInsertRows( QModelIndex(), tagrow, tagrow );
                        tagNode = new TreeNode( TreeNode::TagNode, tag, m_rootNode );
                        m_rootNode->appendChild( tagNode );
                        m_tagNodes[ tag ] = tagNode;
                        endInsertRows();
                }

                QModelIndex tagIndex = createIndex( tagNode->row(), 0, tagNode );
                int feedrow = tagNode->count();
                beginInsertRows( tagIndex, feedrow, feedrow );
                TreeNode *feedNode = new TreeNode( TreeNode::FeedNode, col.id(), tagNode );
                tagNode->appendChild( feedNode );
                m_feedNodes[ col.id() ].append( feedNode );
                endInsertRows();

                // HACK: update the tag node to trigger updates
                // in FeedlistFilterProxyModel
                emit dataChanged( tagIndex, tagIndex );
        }
}

#include "feedlistmodel.moc"
