/*
    Copyright (C) 2008    Dmitry Ivanov <vonami@gmail.com>
    Part of the code borrowed from kdepimlibs/akonadi/standardactionmanager.cpp

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "krssinterface.h"
#include "standardactionmanager.h"
#include "feedlistmodel.h"
#include "subscriptionsmodel.h"
#include "feedpropertiesdialog.h"
#include "../libkrss/feedpropertiescollectionattribute.h"
#include "../libkrss/tagscollectionattribute.h"
#include "../libkrss/subscriptionlabelscollectionattribute.h"

#include <akonadi/collectionmodel.h>
#include <akonadi/collectioncreatejob.h>
#include <akonadi/collectionmodifyjob.h>
#include <akonadi/collectiondeletejob.h>
#include <akonadi/agentmanager.h>
#include <KAction>
#include <KActionCollection>
#include <KDebug>
#include <KLocale>
#include <KInputDialog>
#include <KMessageBox>

#include <QItemSelectionModel>
#include <QTreeView>

#include <boost/static_assert.hpp>

static const struct {
        const char *name;
        const char *label;
        const char *icon;
        int shortcut;
        const char* slot;
} actionData[] = {
        { "krssreader_feed_create", I18N_NOOP( "&New Feed..." ), "folder-new", 0, SLOT( slotCreateFeed() ) },
        { "krssreader_feed_sync", I18N_NOOP( "&Fetch Feed" ), "view-refresh", Qt::Key_F5, SLOT( slotSynchronizeFeed() ) },
        { "krssreader_feed_properties", I18N_NOOP( "Feed &Properties..." ), "configure", 0, SLOT( slotFeedProperties() ) },
        { "krssreader_feed_delete", I18N_NOOP( "&Delete Feed" ), "edit-delete", 0, SLOT( slotDeleteFeed() ) },
        { "krssreader_feed_attachnewtag", I18N_NOOP( "&Attach a new tag" ), "", 0, SLOT( slotAttachNewTag() ) },
        { "krssreader_feed_detachtag", I18N_NOOP( "&Detach the tag" ), "", 0, SLOT( slotDetachTag() ) },
        { "krssreader_manage_subscriptions", I18N_NOOP( "&Manage subscriptions" ), "", 0, SLOT( slotManageSubscriptions() ) }
};
static const int numActionData = sizeof (actionData) / sizeof (*actionData);

BOOST_STATIC_ASSERT( numActionData == StandardActionManager::LastType );

StandardActionManager::StandardActionManager( KActionCollection *actionCollection, QWidget *parent )
        : m_actionCollection( actionCollection ), m_parentWidget( parent )
{
        m_actions.fill( 0, StandardActionManager::LastType );
}

StandardActionManager::~StandardActionManager()
{
}

void StandardActionManager::setFeedSelectionModel( QItemSelectionModel *feedSelectionModel )
{
        m_feedSelectionModel = feedSelectionModel;
        connect( m_feedSelectionModel, SIGNAL( selectionChanged( QItemSelection, QItemSelection ) ),
                 this, SLOT( updateActions() ) );
}


void StandardActionManager::setSubscriptionLabel( const QString &subscriptionLabel )
{
        m_subscriptionLabel = subscriptionLabel;
}

KAction* StandardActionManager::action( StandardActionManager::Type type )
{
        Q_ASSERT( type >= 0 && type < StandardActionManager::LastType );
        Q_ASSERT( actionData[ type ].name );

        if ( m_actions[ type ] )
                return m_actions[ type ];

        KAction *action = new KAction( m_parentWidget );
        if ( actionData[ type ].label )
                action->setText( i18n( actionData[ type ].label ) );

        if ( actionData[ type ].icon )
                action->setIcon( KIcon( QString::fromLatin1( actionData[ type ].icon ) ) );

        action->setShortcut( actionData[ type ].shortcut );

        if ( actionData[ type ].slot )
                connect( action, SIGNAL( triggered() ),
                         this, actionData[ type ].slot );

        m_actionCollection->addAction( QString::fromLatin1( actionData[ type ].name ), action );
        m_actions[ type ] = action;
        updateActions();
        return action;
}

void StandardActionManager::createAllActions()
{
        for ( int i = 0; i < LastType; ++i )
                action( (Type)i );
}

void StandardActionManager::enableAction( StandardActionManager::Type type, bool enable )
{
        Q_ASSERT( type >= 0 && type < StandardActionManager::LastType );
        if ( m_actions[ type ] )
                m_actions[ type ]->setEnabled( enable );
}

void StandardActionManager::updateActions()
{
        // FIXME: only single selections
        QModelIndex selectedIndex;
        if ( m_feedSelectionModel && m_feedSelectionModel->hasSelection() )
                selectedIndex = m_feedSelectionModel->selectedRows().first();

        if ( !selectedIndex.isValid() )
                return;

        // FIXME: better to query the type of the node
        const Akonadi::Collection col = selectedIndex.data( Akonadi::CollectionModel::CollectionRole ).value<Akonadi::Collection>();
        enableAction( StandardActionManager::CreateFeed, true );
        enableAction( StandardActionManager::SynchronizeFeed, col.isValid() );
        enableAction( StandardActionManager::FeedProperties, col.isValid() );
        enableAction( StandardActionManager::DeleteFeed, col.isValid() &&
                      ( col.rights() & Akonadi::Collection::CanDeleteCollection ) );
        enableAction( StandardActionManager::AttachNewTag, col.isValid() );
        enableAction( StandardActionManager::DetachTag, col.isValid() && col.hasAttribute( "Tags" ) &&
                      !col.attribute<TagsCollectionAttribute>()->tags().isEmpty() );
        enableAction( StandardActionManager::ManageSubscriptions, true );
}

void StandardActionManager::slotCreateFeed()
{
        const QString url = KInputDialog::getText( i18nc( "@title:window", "New Source"),
                                                  i18nc( "@label:textbox, name of a thing", "Feed URL"),
                                                  QString(), 0, m_parentWidget );

        if ( url.isEmpty() )
                return;

        //org::kde::krss *krssInterface = new org::kde::krss( "org.kde.krss", "/", QDBusConnection::sessionBus(), this );
        //krssInterface->addFeed( url, "krssreader" );

        // this stuff is broken
        /*const FeedlistModel *sourceModel = qobject_cast<const FeedlistModel*>( m_feedSelectionModel->model() );
        Akonadi::Collection::Id parentId = sourceModel->rootCollection().id();
        Akonadi::Collection col;
        col.setName( url );
        col.setParent( parentId );
        col.setContentMimeTypes( QStringList( "application/rss+xml" ) );
        FeedPropertiesCollectionAttribute *feedProperties = new FeedPropertiesCollectionAttribute();
        feedProperties->setXmlUrl( url );
        col.addAttribute( feedProperties );
        Akonadi::CollectionCreateJob *job = new Akonadi::CollectionCreateJob( col );
        connect( job, SIGNAL( result( KJob* ) ),
                 this, SLOT( collectionCreationResult( KJob* ) ) );*/
}

void StandardActionManager::slotSynchronizeFeed()
{
        // FIXME: only single selections
        QModelIndex selectedIndex;
        if ( m_feedSelectionModel && m_feedSelectionModel->hasSelection() )
                selectedIndex = m_feedSelectionModel->selectedRows().first();

        if ( !selectedIndex.isValid() )
                return;

        const Akonadi::Collection col = selectedIndex.data( Akonadi::CollectionModel::CollectionRole ).value<Akonadi::Collection>();
        Akonadi::AgentManager::self()->synchronizeCollection( col );
}

void StandardActionManager::slotFeedProperties()
{
        QModelIndex selectedIndex;
        if ( m_feedSelectionModel && m_feedSelectionModel->hasSelection() )
                selectedIndex = m_feedSelectionModel->selectedRows().first();

        if ( !selectedIndex.isValid() )
                return;

        Akonadi::Collection col = selectedIndex.data( Akonadi::CollectionModel::CollectionRole ).value<Akonadi::Collection>();

        FeedPropertiesDialog *dialog = new FeedPropertiesDialog();
        dialog->setFeedName( col.name() );
        if ( col.hasAttribute( "FeedProperties" ) ) {
                FeedPropertiesCollectionAttribute *props = col.attribute<FeedPropertiesCollectionAttribute>();
                dialog->setUrl( props->xmlUrl() );
        }

        if ( dialog->exec() == QDialog::Accepted ) {
                // set feed's name
                col.setName( dialog->feedName() );

                // set feed's properties
                FeedPropertiesCollectionAttribute *props;
                if ( col.hasAttribute( "FeedProperties" ) ) {
                        props = col.attribute<FeedPropertiesCollectionAttribute>();
                }
                else {
                        props = new FeedPropertiesCollectionAttribute();
                        col.addAttribute( props );
                }
                props->setXmlUrl( dialog->url() );

                // save the collection
                Akonadi::CollectionModifyJob *job = new Akonadi::CollectionModifyJob( col, this );
                connect( job, SIGNAL( result( KJob* ) ),
                         this, SLOT( collectionModifyResult( KJob* ) ) );
        }

        delete dialog;
}

void StandardActionManager::slotDeleteFeed()
{
        // FIXME: only single selections
        QModelIndex selectedIndex;
        if ( m_feedSelectionModel && m_feedSelectionModel->hasSelection() )
                selectedIndex = m_feedSelectionModel->selectedRows().first();

        if ( !selectedIndex.isValid() )
                return;

        const Akonadi::Collection col = selectedIndex.data( Akonadi::CollectionModel::CollectionRole ).value<Akonadi::Collection>();
        Akonadi::CollectionDeleteJob *job = new Akonadi::CollectionDeleteJob( col, this );
        connect( job, SIGNAL( result( KJob* ) ),
                  this, SLOT( collectionDeletionResult( KJob* ) ) );
}

void StandardActionManager::slotAttachNewTag()
{
        QModelIndex selectedIndex;
        if ( m_feedSelectionModel && m_feedSelectionModel->hasSelection() )
                selectedIndex = m_feedSelectionModel->selectedRows().first();

        if ( !selectedIndex.isValid() )
                return;

        const QString tag = KInputDialog::getText( i18nc( "@title:window", "Add a new tag to this feed"),
                                                  i18nc( "@label:textbox, name of a thing", "New tag"),
                                                  QString(), 0, m_parentWidget );

        if ( tag.isEmpty() )
                return;

        Akonadi::Collection col = selectedIndex.data( Akonadi::CollectionModel::CollectionRole ).value<Akonadi::Collection>();
        if ( col.hasAttribute( "Tags" ) ) {
                col.attribute<TagsCollectionAttribute>()->addTag( tag );
        }
        else {
                TagsCollectionAttribute *tags = new TagsCollectionAttribute( QStringList( tag ) );
                col.addAttribute( tags );
        }

        Akonadi::CollectionModifyJob *job = new Akonadi::CollectionModifyJob( col, this );
        connect( job, SIGNAL( result( KJob* ) ),
                  this, SLOT( collectionModifyResult( KJob* ) ) );
}

void StandardActionManager::slotDetachTag()
{
        QModelIndex selectedIndex;
        if ( m_feedSelectionModel && m_feedSelectionModel->hasSelection() )
                selectedIndex = m_feedSelectionModel->selectedRows().first();

        if ( !selectedIndex.isValid() )
                return;

        Akonadi::Collection col = selectedIndex.data( Akonadi::CollectionModel::CollectionRole ).value<Akonadi::Collection>();
        QString tag = selectedIndex.parent().data( Qt::DisplayRole ).toString();

        // 'tags' attribute exists, checked in enabledActions()
        col.attribute<TagsCollectionAttribute>()->removeTag( tag );

        Akonadi::CollectionModifyJob *job = new Akonadi::CollectionModifyJob( col, this );
        connect( job, SIGNAL( result( KJob* ) ),
                  this, SLOT( collectionModifyResult( KJob* ) ) );
}

void StandardActionManager::slotManageSubscriptions()
{
        KDialog *manager = new KDialog( m_parentWidget );
        manager->setCaption( i18n( "Manage RSS subscriptions" ) );
        manager->setButtons( KDialog::Ok | KDialog::Cancel );

        Akonadi::CollectionModel *fullModel = new Akonadi::CollectionModel( manager );
        SubscriptionsModel *subscriptionsModel = new SubscriptionsModel( m_subscriptionLabel, manager );
        subscriptionsModel->setSourceModel( fullModel );
        QTreeView *subscriptionsView = new QTreeView( manager );
        subscriptionsView->setModel( subscriptionsModel );
        manager->setMainWidget( subscriptionsView );

        if ( manager->exec() == QDialog::Accepted ) {
                kDebug() << "Old subscriptions:" << subscriptionsModel->oldSubscriptions();
                kDebug() << "New subscriptions:" << subscriptionsModel->newSubscriptions();

                QSet<Akonadi::Collection::Id> toRemove = subscriptionsModel->oldSubscriptions().
                                                         subtract( subscriptionsModel->newSubscriptions() );
                QSet<Akonadi::Collection::Id> toAdd = subscriptionsModel->newSubscriptions().
                                                         subtract( subscriptionsModel->oldSubscriptions() );
                kDebug() << "to remove:" << toRemove;
                kDebug() << "to add:" << toAdd;

                Q_FOREACH( const Akonadi::Collection::Id &colid, toRemove ) {
                        Akonadi::Collection col = subscriptionsModel->collectionForId( colid );
                        SubscriptionLabelsCollectionAttribute *attr = col.attribute<SubscriptionLabelsCollectionAttribute>(
                                                                                 Akonadi::Entity::AddIfMissing );

                        kDebug() << "Unsubscribing from col:" << col.name();
                        attr->removeSubscriptionLabel( m_subscriptionLabel );
                        Akonadi::CollectionModifyJob *job = new Akonadi::CollectionModifyJob( col, this );
                        connect( job, SIGNAL( result( KJob* ) ),
                                 this, SLOT( collectionModifyResult( KJob* ) ) );
                }

                Q_FOREACH( const Akonadi::Collection::Id &colid, toAdd ) {
                        Akonadi::Collection col = subscriptionsModel->collectionForId( colid );
                        SubscriptionLabelsCollectionAttribute *attr = col.attribute<SubscriptionLabelsCollectionAttribute>(
                                                                                 Akonadi::Entity::AddIfMissing );

                        kDebug() << "Subscribing to col:" << col.name();
                        attr->addSubscriptionLabel( m_subscriptionLabel );
                        Akonadi::CollectionModifyJob *job = new Akonadi::CollectionModifyJob( col, this );
                        connect( job, SIGNAL( result( KJob* ) ),
                                 this, SLOT( collectionModifyResult( KJob* ) ) );
                }
        }

        delete manager;
}

void StandardActionManager::collectionCreationResult( KJob* job )
{
        if ( job->error() ) {
                KMessageBox::error( m_parentWidget, i18n("Could not add a new feed: %1", job->errorString() ),
                                    i18n( "Feed creation failed" ) );
        }
}

void StandardActionManager::collectionDeletionResult( KJob* job )
{
        if ( job->error() ) {
                KMessageBox::error( m_parentWidget, i18n("Could not delete the feed: %1", job->errorString() ),
                                    i18n( "Feed deleteion failed" ) );
        }
}

void StandardActionManager::collectionModifyResult( KJob* job )
{
        if ( job->error() ) {
                KMessageBox::error( m_parentWidget, i18n("Could not change feed properties: %1", job->errorString() ),
                                    i18n( "Feed properties changing failed" ) );
        }
}

#include "standardactionmanager.moc"
