/*
    Copyright (C) 2008    Dmitry Ivanov <vonami@gmail.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "krssadaptor.h"
#include "rssresource.h"
#include "../libkrss/tagscollectionattribute.h"
#include "../libkrss/subscriptionlabelscollectionattribute.h"
#include "../libkrss/feedpropertiescollectionattribute.h"

#include <kfeed/person.h>
#include <kfeed/enclosure.h>
#include <kfeed/category.h>
#include <akonadi/collectionfetchjob.h>
#include <akonadi/collectioncreatejob.h>
#include <akonadi/collectionmodifyjob.h>
#include <akonadi/collectiondeletejob.h>
#include <akonadi/attributefactory.h>
#include <akonadi/changerecorder.h>

#include <KDebug>
#include <KFileDialog>
#include <KDateTime>
#include <KLocale>
#include <KRandom>

#include <QDBusConnection>

using namespace Akonadi;

RssResource::RssResource(const QString &id)
        :ResourceBase(id), m_fetchTries(0)
{
        AttributeFactory::registerAttribute<TagsCollectionAttribute>();
        AttributeFactory::registerAttribute<SubscriptionLabelsCollectionAttribute>();
        AttributeFactory::registerAttribute<FeedPropertiesCollectionAttribute>();

        new KrssAdaptor( this );
        if ( !QDBusConnection::sessionBus().registerService( "org.kde.krss" ) ) {
                kWarning() << "Couldn't register a D-Bus service org.kde.krss";
                kWarning() << QDBusConnection::sessionBus().lastError().message();
        }

        changeRecorder()->fetchCollection( true );
}

void RssResource::triggerBug()
{
        Akonadi::Collection col = m_collections[ "bug" ];
        col.setName( "NEW NAME" );
        m_changedCollections << col;
        synchronizeCollectionTree();
}

bool RssResource::retrieveItem(const Akonadi::Item &item, const QSet<QByteArray> &parts)
{
        Q_UNUSED(item)
        Q_UNUSED(parts)

        // what should i do?
        // no way to retrieve a single item from an RSS feed other than
        // refetching the whole feed and looking up the item in the feed
        // (actually, the feed might not contain this item anymore)
        return false;
}

void RssResource::configure(WId windowId)
{
        // note: this WId trick doesn't work
        QString path = KFileDialog::getOpenFileNameWId( KUrl::fromPath( QDir::homePath() ),
                                                 "*.opml|" + i18n("OPML Document (*.opml)"),
                                                 windowId, i18n("Select an OPML Document"));
        if ( !path.isEmpty() ) {
                Collection root;
                root.setParent( Collection::root() );
                root.setRemoteId( path );
                root.setName( "RSS subscriptions" );
                root.setContentMimeTypes( QStringList( Collection::mimeType() ) );

                m_rootCollection = root;
                m_changedCollections << root;

                loadOpml( path );
                synchronize();
        }
}

void RssResource::retrieveCollections()
{
        kDebug() << "Before collectionsRetrievedIncremental()";
        dump();

        collectionsRetrievedIncremental( m_changedCollections, m_removedCollections );

        // process changed collections
        Q_FOREACH( const Akonadi::Collection &col, m_changedCollections ) {
                m_collections[ col.remoteId() ] = col;
        }

        // process removed collections
        Q_FOREACH( const Akonadi::Collection &col, m_removedCollections ) {
                m_collections.remove( col.remoteId() );
        }

        // cleanup
        m_changedCollections.clear();
        m_removedCollections.clear();
}

void RssResource::retrieveItems(const Akonadi::Collection &col)
{
        kDebug() << "Retrieving items for" << col.name();

        if ( !col.isValid() ) {
                kDebug() << "Collection is not valid";
                return;
        }

        if ( col.contentMimeTypes().contains(Collection::mimeType()) ) {
                kDebug() << "Collection" << col.name() << "is a folder";
                return;
        }

        if ( col.hasAttribute( "FeedProperties" ) ) {
                fetchFeedContent( col.attribute<FeedPropertiesCollectionAttribute>()->xmlUrl() );
        }
        else {
                kDebug() << "Collection doesn't have \'FeedProperties\' attribute set";
                Item::List itemList;
                itemsRetrieved(itemList);
        }
}

void RssResource::fetchFeedContent( const QString &url )
{
        Syndication::Loader *loader = Syndication::Loader::create( this, SLOT( feedContentFetched( Syndication::Loader*,
                                                                  Syndication::FeedPtr,
                                                                  Syndication::ErrorCode ) ) );

        kDebug() << "Url:" << url;
        loader->loadFrom( url );
}

void RssResource::fetchFeedProperties( const QString &url )
{
        Syndication::Loader *loader = Syndication::Loader::create( this, SLOT( feedPropertiesFetched( Syndication::Loader*,
                                                                  Syndication::FeedPtr,
                                                                  Syndication::ErrorCode ) ) );

        kDebug() << "Url:" << url;
        loader->loadFrom( url );
}

void RssResource::feedPropertiesFetched( Syndication::Loader *loader, Syndication::FeedPtr feed, Syndication::ErrorCode status )
{
        kDebug() << "Status:" << status;

        if ( ( status == Syndication::InvalidXml ) && (m_fetchTries < 3 ) && ( loader->discoveredFeedURL().isValid() ) ) {
                kDebug() << "Refetching by discovered feed url";
                QString newUrl = loader->discoveredFeedURL().url();
                m_changedCollections[ 0 ].setName( newUrl );
                m_changedCollections[ 0 ].attribute<FeedPropertiesCollectionAttribute>()->setXmlUrl( newUrl );
                m_fetchTries++;
                fetchFeedProperties( newUrl );
                return;
        }

        if ( status == Syndication::Success ) {
                // set the corresponding collection attributes
                m_changedCollections[ 0 ].attribute<FeedPropertiesCollectionAttribute>()->setName( feed->title() );
                m_changedCollections[ 0 ].attribute<FeedPropertiesCollectionAttribute>()->setHtmlUrl( feed->link() );
                m_changedCollections[ 0 ].attribute<FeedPropertiesCollectionAttribute>()->setFeedType( "rss" );
                m_changedCollections[ 0 ].attribute<FeedPropertiesCollectionAttribute>()->setDescription( feed->description() );
        }

        m_fetchTries = 0;

        // commit the new collection
        Akonadi::CollectionCreateJob *job = new Akonadi::CollectionCreateJob( m_changedCollections[ 0 ] );
        connect( job, SIGNAL( result( KJob* ) ),
                 this, SLOT( slotCollectionCreated( KJob* ) ) );
}

void RssResource::feedContentFetched( Syndication::Loader *loader, Syndication::FeedPtr feed, Syndication::ErrorCode status )
{
        Q_UNUSED( loader )

        kDebug() << "Status:" << status;

        Item::List itemList;
        if ( status != Syndication::Success ) {
                kDebug() << "Failed to load feed from"
                        << currentCollection().attribute<FeedPropertiesCollectionAttribute>()->xmlUrl();
                itemsRetrieved( itemList );
                return;
        }

        Q_FOREACH( const Syndication::ItemPtr &it, feed->items() ) {
                Item item;
                item.setRemoteId( it->id() );
                item.setMimeType( "application/rss+xml" );
                item.setPayload<KFeed::Item>( convertToKFeedItem( it ) );
                itemList.append( item );
        }

        itemsRetrieved( itemList );
}

KFeed::Item RssResource::convertToKFeedItem(const Syndication::ItemPtr &syndItem)
{
                KFeed::Item kfeedItem;

                kfeedItem.setStatus(KFeed::New);             // shouldn't it be KFeed::New | KFeed::Unread?
                //kfeedItem.setHash()                        // do i need to calc the hash value?
                kfeedItem.setIdIsHash(syndItem->id().startsWith("hash:"));
                kfeedItem.setSourceFeedId(currentCollection().remoteId().toInt());
                kfeedItem.setTitle(syndItem->title());
                kfeedItem.setLink(syndItem->link());
                kfeedItem.setDescription(syndItem->description());
                kfeedItem.setContent(syndItem->content());
                KDateTime dt;
                if ( syndItem->datePublished() > 0 ) {
                        dt.setTime_t(syndItem->datePublished());
                        kfeedItem.setDatePublished(dt);
                }
                else {
                        kfeedItem.setDatePublished(KDateTime::currentLocalDateTime());
                }
                dt.setTime_t(syndItem->dateUpdated());
                kfeedItem.setDateUpdated(dt);
                kfeedItem.setId(syndItem->id());
                kfeedItem.setLanguage(syndItem->language());
                kfeedItem.setCommentsCount(syndItem->commentsCount());
                kfeedItem.setCommentsLink(syndItem->commentsLink());
                kfeedItem.setCommentsFeed(syndItem->commentsFeed());
                kfeedItem.setCommentPostUri(syndItem->commentPostUri());

                // Authors
                QList<KFeed::Person> kfeedPersons;
                Q_FOREACH(const Syndication::PersonPtr &pers, syndItem->authors()) {
                        if ( !pers->isNull() ) {
                                KFeed::Person kfeedPerson;
                                kfeedPerson.setName(pers->name());
                                kfeedPerson.setUri(pers->uri());
                                kfeedPerson.setEmail(pers->email());
                                kfeedPersons << kfeedPerson;
                        }
                }
                kfeedItem.setAuthors(kfeedPersons);

                // Enclosures
                QList<KFeed::Enclosure> kfeedEnclosures;
                Q_FOREACH(const Syndication::EnclosurePtr &encl, syndItem->enclosures()) {
                        if ( !encl->isNull() ) {
                                KFeed::Enclosure kfeedEnclosure;
                                kfeedEnclosure.setUrl(encl->url());
                                kfeedEnclosure.setTitle(encl->title());
                                kfeedEnclosure.setType(encl->type());
                                kfeedEnclosure.setLength(encl->length());
                                kfeedEnclosure.setDuration(encl->duration());
                                kfeedEnclosures << kfeedEnclosure;
                        }
                }
                kfeedItem.setEnclosures(kfeedEnclosures);

                // Categories
                QList<KFeed::Category> kfeedCategories;
                Q_FOREACH(const Syndication::CategoryPtr &cat, syndItem->categories()) {
                        if ( !cat->isNull() ) {
                                KFeed::Category kfeedCategory;
                                kfeedCategory.setTerm(cat->term());
                                kfeedCategory.setScheme(cat->scheme());
                                kfeedCategory.setLabel(cat->label());
                                kfeedCategories << kfeedCategory;
                        }
                }
                kfeedItem.setCategories(kfeedCategories);

                // TODO: custom properties

                return kfeedItem;
}

void RssResource::dump()
{
        kDebug() << "**************************************************";
        kDebug() << "Current collections:";
        QHashIterator<QString, Akonadi::Collection> it(m_collections);
        while ( it.hasNext() ) {
                it.next();
                kDebug() << "Name:" << it.value().name() << ", remote id:" << it.value().remoteId()
                         << ", local id:" << it.value().id();
        }

        kDebug() << "Changed collections:";
        Q_FOREACH( const Akonadi::Collection &col, m_changedCollections ) {
                kDebug() << "Name:" << col.name() << ", remote id:" << col.remoteId() << "local id:" << col.id();
        }

        kDebug() << "Removed collections:";
        Q_FOREACH( const Akonadi::Collection &col, m_removedCollections ) {
                kDebug() << "Name:" << col.name() << ", remote id:" << col.remoteId() << "local id:" << col.id();
        }
        kDebug() << "**************************************************";
}

bool RssResource::loadOpml( const QString &path )
{
        QFile file( path );
        if ( !file.open(QFile::ReadOnly | QFile::Text) ) {
                kWarning() << "Couldn't open the OPML document";
                return false;
        }

        m_reader.setDevice(&file);

        while ( !m_reader.atEnd() ) {
                m_reader.readNext();

                if ( m_reader.isStartElement() ) {
                        if ( m_reader.name().toString().toLower() == "opml" ) {
                                kDebug() << "OPML version" << m_reader.attributes().value( "version" ).toString();
                                readOpml();
                        }
                        else {
                                m_reader.raiseError( "The file is not an OPML document." );
                        }
                }
        }

        file.close();
        return !m_reader.error();
}

void RssResource::readUnknownElement()
{
        Q_ASSERT(m_reader.isStartElement());

        while ( !m_reader.atEnd() ) {
                m_reader.readNext();

                if ( m_reader.isEndElement() )
                        break;

                if ( m_reader.isStartElement() )
                        readUnknownElement();
        }
}

void RssResource::readOpml()
{
        Q_ASSERT( m_reader.isStartElement() && m_reader.name().toString().toLower() == "opml" );

        while ( !m_reader.atEnd() ) {
                m_reader.readNext();

                if ( m_reader.isEndElement() )
                        break;

                if ( m_reader.isStartElement() ) {
                        if ( m_reader.name().toString().toLower() == "head" ) {
                                readUnknownElement();
                        }
                        else if ( m_reader.name().toString().toLower() == "body" ) {
                                readBody();
                        }
                        else {
                                readUnknownElement();
                        }
                }
        }
}

void RssResource::readBody()
{
        Q_ASSERT( m_reader.isStartElement() && m_reader.name().toString().toLower() == "body" );

        while ( !m_reader.atEnd() ) {
                m_reader.readNext();

                if ( m_reader.isEndElement() )
                        break;

                if ( m_reader.isStartElement() ) {
                        if ( m_reader.name().toString().toLower() == "outline" ) {
                                QStringList tags;
                                readOutline( tags );
                        }
                        else {
                                readUnknownElement();
                        }
                }
        }
}

static QStringRef attributeValue( const QXmlStreamAttributes &attributes, const QString &name )
{
        Q_FOREACH( const QXmlStreamAttribute &attr, attributes ) {
                if ( attr.name().toString().toLower() == name ) {
                        return attr.value();
                }
        }

        return QStringRef();
}

void RssResource::readOutline( QStringList &tags )
{
        Q_ASSERT( m_reader.isStartElement() && m_reader.name().toString().toLower() == "outline" );

        bool isFolder = false;
        QString xmlUrl = attributeValue( m_reader.attributes(), "xmlurl" ).toString();

        if ( xmlUrl.isEmpty() ) {
                if ( !attributeValue( m_reader.attributes(), "text" ).isEmpty() ) {
                        // this is an Akregator's folder
                        isFolder = true;
                        tags << attributeValue( m_reader.attributes(), "text" ).toString();
                        kDebug() << "Tags:" << tags;
                }
                else {
                        kDebug() << "Encountered an empty outline";
                        // TODO: dump the rest of the block
                }
        }
        else {
                // this is a feed
                kDebug() << "Feed:" << xmlUrl;
                isFolder = false;
                Collection col;
                if ( xmlUrl == "http://planetkde.org/rss20.xml" ) {
                        col.setName( "OLD NAME" );
                        col.setRemoteId( "bug" );
                }
                else {
                        col.setName( xmlUrl );
                        col.setRemoteId( KRandom::randomString( 10 ) );
                }
                col.setParent( m_rootCollection );
                col.setContentMimeTypes( QStringList( "application/rss+xml" ) );

                // set feed properties
                FeedPropertiesCollectionAttribute *props = new FeedPropertiesCollectionAttribute();
                props->setName( attributeValue( m_reader.attributes(), "title" ).toString() );
                if ( props->name().isEmpty() ) {
                        props->setName( attributeValue( m_reader.attributes(), "text" ).toString() );
                }
                if ( props->name().isEmpty() ) {
                        props->setName( xmlUrl );
                }
                props->setXmlUrl( xmlUrl );
                props->setHtmlUrl( attributeValue( m_reader.attributes(), "htmlurl" ).toString() );
                props->setDescription( attributeValue( m_reader.attributes(), "description" ).toString() );
                props->setFeedType( attributeValue( m_reader.attributes(), "type" ).toString() );
                if ( props->feedType().isEmpty() ) {
                        props->setFeedType( "rss" );
                }
                col.addAttribute( props );

                // set feed tags
                kDebug() << "Setting tags:" << tags;
                TagsCollectionAttribute *feedTags = new TagsCollectionAttribute( tags );
                col.addAttribute( feedTags );
                m_changedCollections << col;
        }

        while ( !m_reader.atEnd() ) {
                m_reader.readNext();

                if ( m_reader.isEndElement() )
                        break;

                if ( m_reader.isStartElement() ) {
                        if ( m_reader.name().toString().toLower() == "outline" && isFolder )
                                readOutline( tags );
                        else
                                readUnknownElement();
                }
        }

        // once we are back from recursion remove the added tag
        // from the top of the list
        if ( isFolder )
                tags.removeLast();
}

AKONADI_RESOURCE_MAIN(RssResource)

#include "rssresource.moc"
